package infomat.algorithms.grouping;

import infomat.evaluation.IntrinsicGroupEvaluation;
import infomat.vectorspace.*;
import infomat.vectorspace.similarity.*;
import mro.util.*;
import mro.util.properties.*;

import java.util.Vector; 
import java.util.Map.Entry; 
import java.util.HashMap; 
import java.util.HashSet; 
import java.util.Set; 


/**
 * Top-Down-Splitter. Basic class for repeatedly splitting bad clusters.
 *      
 * @version          080415
 * @author           Magnus Rosell
 */
public class TopDownSplitter //extends GroupingAlgorithm 
    implements NeedsISimilarity{ 

    public static int topDownClusteringNumber = 1;
	
    public mro.util.properties.Properties getDefaultProperties(){

	mro.util.properties.Properties properties = 
	    new mro.util.properties.Properties("Top Down Clusterer Properties", CurrentClassGetter.getClassName());
		
	try{

	    Property p = new Property("Number of clusters", 5, "Number of clusers");
	    properties.addProperty(p);

	    p = new Property("Split into", 2, "Split each cluster in");
	    properties.addProperty(p);

	    p = new Property("Split selection", 0, "Which cluster to split");
	    PropertyValueDescription pvd = 
		new PropertyValueDescription(0, "Largest cluster");
	    p.addValueDescription(pvd);
	    properties.addProperty(p);

	    p = new Property("Number of split tries", 10, "Number of times each split is tried");
	    properties.addProperty(p);

	    p = new Property("Split evaluation", 0, "How to decide which split is the best.");
	    pvd = new PropertyValueDescription(0, "Highest weighted average self similarity.");
	    p.addValueDescription(pvd);
	    properties.addProperty(p);


	}catch(PropertyNameConflictException pnce){
	    // Won't happen here
	}
	return properties;
    }


    private ISimilarity similarity;
    private IntrinsicGroupEvaluation intrinsicGroupEvaluation;
    protected Properties properties;
    protected UniqueIdentifier identifier;

    protected boolean printGroupSizesInIterations = true;

	
    public TopDownSplitter(){
	this.properties = getDefaultProperties();
    }

    public TopDownSplitter(mro.util.properties.Properties properties){
	this.properties=properties;
    }

	
    public void setSimilarity(ISimilarity similarity){
	this.similarity = similarity;
	intrinsicGroupEvaluation = new IntrinsicGroupEvaluation(similarity);
    }


    /*
    public IObjectGrouping cluster(IObjectSet set){
	return new IObjectGrouping("Dummy Grouping");
    }

    public IObjectGrouping cluster(IObjectGroup iogroup){
	return new IObjectGrouping("Dummy Grouping");
    }
    */


    public IObjectGrouping cluster(IObjectGroup iogroup, GroupingAlgorithm alg){
		
	int numOfClusters = (Integer) properties.getPropertyValue("Number of clusters");
	int splitInto = (Integer) properties.getPropertyValue("Split into");

	int splitSelection = (Integer) properties.getPropertyValue("Split selection");
	int numberOfTries = (Integer) properties.getPropertyValue("Number of split tries");
	int splitEvaluation = (Integer) properties.getPropertyValue("Split evaluation");

	try{
	    alg.setProperty("Number of clusters", splitInto);
	}catch(PropertyNotFoundException pnfe){
	    // Wrong type of algorithm
	    return null;
	}

	
	Vector<IObjectGroupNode> currentNodes = 
	    new Vector<IObjectGroupNode>();
	IObjectGroupNode root = new IObjectGroupNode(iogroup, null);
	currentNodes.add(root);
	
	int clusterNumber=1;
	while(currentNodes.size() < numOfClusters){
	    //	    IObjectGroupNode worst = removeWorstNode(currentNodes);
	    Vector splitVector = removeWorstNode(currentNodes);
	    IObjectGroupNode worst = (IObjectGroupNode)
		splitVector.elementAt(1);

	    IObjectGroup worstGroup = worst.getIObjectGroup();
	    Vector<IObjectGroup> groups = 
		split(worstGroup, alg, numberOfTries, splitEvaluation);

	    currentNodes = new Vector<IObjectGroupNode>();
	    currentNodes.addAll((Vector<IObjectGroupNode>) 
				splitVector.elementAt(0));
	    for(int i=0; i < groups.size(); i++){
		IObjectGroup group = groups.elementAt(i);
		if(group.numberOfIObjects()>0){
		    group.setName("Cluster "+clusterNumber);
		    clusterNumber++;
		    IObjectGroupNode newNode = 
			new IObjectGroupNode(group, worst);
		    worst.addChild(newNode);
		    currentNodes.add(newNode);
		}
	    }
	    currentNodes.addAll((Vector<IObjectGroupNode>) 
				splitVector.elementAt(2));

	    if(printGroupSizesInIterations)
		printGroupSizes(currentNodes);
	}



	String name = "Top Down Splitter Clustering "+topDownClusteringNumber;
	topDownClusteringNumber++;
	HierarchicalIObjectGrouping hioging = 
	    new HierarchicalIObjectGrouping(name);
	Vector<IObjectGroupNode> rootNodeVec = new Vector<IObjectGroupNode>();
	rootNodeVec.add(root);
	hioging.setRootIObjectGroupNodes(rootNodeVec);
	hioging.setBottomIObjectGroupNodes(currentNodes);

	hioging.setGroupingToBottomGrouping();

	/*
	System.out.println("------------ HIERARCHY --------------");
	hioging.printHierarchy();
	System.out.println("---------- END HIERARCHY ------------");
	*/

	return hioging;
    }


    private Vector<IObjectGroup> split(IObjectGroup worstGroup, GroupingAlgorithm alg, int numberOfTries, int splitEvaluation){

	IObjectGrouping ioging = alg.cluster(worstGroup);
	Vector<IObjectGroup> bestGroups = ioging.getGroups();
	double bestSS = 
	    intrinsicGroupEvaluation.weightedAvgSelfSimilarity(bestGroups);
	for(int i=1; i<numberOfTries; i++){
	    ioging = alg.cluster(worstGroup);
	    Vector<IObjectGroup> groups = ioging.getGroups();
	    double ss = intrinsicGroupEvaluation.weightedAvgSelfSimilarity(groups);
	    if(ss > bestSS){
		bestSS = ss;
		bestGroups = groups;
	    }
	}
	return bestGroups;
    }

    /*
    public IObjectGrouping cluster(IObjectSet ioset, GroupingAlgorithm alg){
	IObjectGrouping grouping = cluster((IObjectGroup) ioset, alg);
	String name = grouping.getName();

	int i=1;
	while(!ioset.setGrouping(grouping)){
	    grouping.setName(name+" ("+i+")");
	    i++;
	}
	return grouping;
    }
    */	

    private Vector
	removeWorstNode(Vector<IObjectGroupNode> currentNodes){
	
	IObjectGroupNode rmGroupNode = currentNodes.elementAt(0);

	int splitSelection = 
	    (Integer) properties.getPropertyValue("Split selection");

	int rmi = 0;
	if(splitSelection==0){ // largest cluster
	    for(int i=0; i < currentNodes.size(); i++){
		if(currentNodes.elementAt(i).getIObjectGroup().numberOfIObjects() > rmGroupNode.getIObjectGroup().numberOfIObjects()){
		    rmGroupNode = currentNodes.elementAt(i);
		    rmi=i;
		}
	    }
	}
	
	Vector ret = new Vector();
	Vector<IObjectGroupNode> tmpVec = new Vector<IObjectGroupNode>();
	for(int i=0; i<rmi; i++)
	    tmpVec.add(currentNodes.elementAt(i));
	ret.add(tmpVec);
	ret.add(rmGroupNode);
	tmpVec = new Vector<IObjectGroupNode>();
	for(int i=rmi+1; i<currentNodes.size(); i++)
	    tmpVec.add(currentNodes.elementAt(i));
	ret.add(tmpVec);

	//	currentNodes.remove(rmGroupNode);

	return ret;
    }

	
	
    /**
     * For debuggin purposes!
     */	
    protected void printGroupSizes(Vector<IObjectGroupNode> nodes){
	for(int i=0; i<nodes.size(); i++){
	    IObjectGroup group = nodes.elementAt(i).getIObjectGroup();
	    System.out.print(group.numberOfIObjects()+" ");
	}
	System.out.println();
    }


    public String getInfoString(){
    	return "TopDownSplitter";
    }

    public String getAlgorithmName(){
    	return "TopDownSplitter";    	
    }
    
    public Vector getPropertyVector(){
    	return properties.getPropertyVector();
    }
    

    public Object getPropertyValue(String name){
    	return properties.getPropertyValue(name);
    }
    public void setProperty(String name, Object value) throws PropertyNotFoundException{
    	properties.setProperty(name, value);
    }

    public mro.util.properties.Properties getProperties(){
    	return properties;
    }
    public void setProperties(mro.util.properties.Properties properties){
    	this.properties=properties;
    }

    public UniqueIdentifier getUniqueIdentifier(){
    	return identifier;
    }

    public void setUniqueIdentifier(UniqueIdentifier uniqueIdentifier){
    	this.identifier=uniqueIdentifier;
    }

}
