package infomat.evaluation;

import java.util.*;
import java.io.*;

import infomat.vectorspace.*;

import mro.util.*;
import mro.util.measures.*;
import mro.util.doublehashmap.DoubleHashMap;


/**
 * For calculating extrinsic quality measures per IObjectGroup.
 * 
 * @version          080424
 * @author           Magnus Rosell
 */
public class ExtrinsicGroupEvaluation {

    private IObjectGrouping evalGrouping;
    private IObjectGrouping refGrouping;
    private DoubleHashMap confusionMatrix;

    public ExtrinsicGroupEvaluation(IObjectGrouping evalGrouping, 
				    IObjectGrouping refGrouping){
	this.evalGrouping=evalGrouping;
	this.refGrouping=refGrouping;
	constructConfusionMatrix(evalGrouping, refGrouping);
    }

    protected ExtrinsicGroupEvaluation(DoubleHashMap confusionMatrix, 
				       IObjectGrouping evalGrouping, 
				       IObjectGrouping refGrouping){
	this.confusionMatrix = confusionMatrix;
	this.evalGrouping=evalGrouping;
	this.refGrouping=refGrouping;
    }

    private void constructConfusionMatrix(IObjectGrouping evalGrouping, IObjectGrouping refGrouping){
	confusionMatrix = new DoubleHashMap();
	Vector<IObjectGroup> g1groups = evalGrouping.getGroups();
	Vector<IObjectGroup> g2groups = refGrouping.getGroups();

	for(int i = 0; i < g1groups.size(); i++){
	    IObjectGroup iog1 = g1groups.get(i);
	    for(int j = 0; j < g2groups.size(); j++){
		IObjectGroup iog2 = g2groups.get(j);
		confusionMatrix.put(iog1, iog2, iog1.numberOfSharedIObjects(iog2));
	    }
	}		
    }

    public HashMap<IObjectGroup, Measure> getGroupPurityMap(){
	HashMap<IObjectGroup, Measure> iogmeasure = new HashMap<IObjectGroup, Measure>();
	Vector<IObjectGroup> vec = evalGrouping.getGroups();
	for(int i = 0; i < vec.size(); i++){
	    IObjectGroup iog = vec.get(i);
	    Measure m = purity(iog);
	    iogmeasure.put(iog, m);

	}

	return iogmeasure;
    }

    public HashMap<IObjectGroup, Measure> getGroupEntropyMap(){
	HashMap<IObjectGroup, Measure> iogmeasure = new HashMap<IObjectGroup, Measure>();
	Vector<IObjectGroup> vec = evalGrouping.getGroups();
	for(int i = 0; i < vec.size(); i++){
	    IObjectGroup iog = vec.get(i);
	    Measure m = entropy(iog);
	    iogmeasure.put(iog, m);
	}

	return iogmeasure;
    }

    public HashMap<IObjectGroup, Measure> getGroupNormEntropyMap(){
	HashMap<IObjectGroup, Measure> iogmeasure = new HashMap<IObjectGroup, Measure>();
	Vector<IObjectGroup> vec = evalGrouping.getGroups();
	for(int i = 0; i < vec.size(); i++){
	    IObjectGroup iog = vec.get(i);
	    Measure m = normEntropy(iog);
	    
	    iogmeasure.put(iog, m);
	}

	return iogmeasure;
    }


    public Measure purity(IObjectGroup iog){
	Vector g2groups = refGrouping.getGroups();
	int count = 0;
	IObjectGroup iogMax = null;
	for(int i = 0; i < g2groups.size(); i++){
	    if((Integer) confusionMatrix.get(iog, g2groups.get(i))>count){
		count = (Integer) confusionMatrix.get(iog, g2groups.get(i));
		iogMax = (IObjectGroup) g2groups.get(i);
	    }

	}

	Measure m = new Measure("Purity", (double) count / iog.numberOfIObjects(), "Purity", iogMax);
	return m;
    }

    public Measure entropy(IObjectGroup iog){
	
	Vector g2groups = refGrouping.getGroups();
	int count = 0;
	double entropy = 0.0;
	for(int i = 0; i < g2groups.size(); i++){
	    if(((Integer) confusionMatrix.get(iog, g2groups.get(i))) > 0){
		count = (Integer) confusionMatrix.get(iog, g2groups.get(i));
		if(count>0){
		    double p = ((double)count)/iog.numberOfIObjects();
		    entropy -= p*MoreMath.log2(p);
		}
	    }
	}

	Measure m = new Measure("Entropy", entropy, "Entropy");
	return m;
    }

    public Measure normEntropy(IObjectGroup iog){
	Vector g2groups = refGrouping.getGroups();
	int count = 0;
	double entropy = 0.0;
	for(int i = 0; i < g2groups.size(); i++){
	    if((Integer) confusionMatrix.get(iog, g2groups.get(i)) > 0){
		count = (Integer) confusionMatrix.get(iog, g2groups.get(i));
		if(count>0){
		    double p = ((double)count)/iog.numberOfIObjects();
		    entropy -= p*MoreMath.log2(p);
		}
	    }
	}
	entropy = (float) (entropy/MoreMath.log2(refGrouping.numberOfIObjectGroups()));

	Measure m = new Measure("Norm Entropy by Number of Groups", entropy, "Normalized Entropy by Number of Groups");
	return m;
    }


    /********************************************************/
    /********************************************************/
    /********************************************************/


    public static HashMap<IObjectGroup, Measure> 
	getGroupPartPurityMap(IObjectGrouping evalGrouping, 
			      IObjectGrouping refGrouping,
			      float[] partSplitValues){

	HashMap<IObjectGroup, Measure> iogmeasure = new HashMap<IObjectGroup, Measure>();
	Vector<IObjectGroup> vec = evalGrouping.getGroups();
	for(int i = 0; i < vec.size(); i++){
	    IObjectGroup iog = vec.get(i);
	    MultiMeasure mm = getGroupPartPurity(iog, refGrouping, partSplitValues);
	    iogmeasure.put(iog, mm);

	}

	return iogmeasure;
    }
	


    public static Measure getGroupPurity(IObjectGroup group, IObjectGrouping grouping){

	HashMap<IObjectGroup, Integer> groupCount = new HashMap<IObjectGroup, Integer>();
	for(int j = 0; j < grouping.getGroups().size(); j++){
	    groupCount.put(grouping.getGroups().get(j), 0);
	}
		
		
	Vector<IObject> iovec = group.getOrdered();
		
	for(int i = 0; i < iovec.size(); i++){

	    IObject io = iovec.get(i);
	    IObjectGroup countGroup = grouping.inWhichIObjectGroup(io);
	    Integer count = groupCount.get(countGroup);
	    count++;
	    groupCount.put(countGroup, count);
			
	}
		
	Set<Map.Entry<IObjectGroup, Integer>> countset = groupCount.entrySet();
	Iterator it = countset.iterator();
	Map.Entry<IObjectGroup, Integer> best = (Map.Entry<IObjectGroup, Integer>) it.next();
	while(it.hasNext()){
	    Map.Entry<IObjectGroup, Integer> e = (Map.Entry<IObjectGroup, Integer>) it.next();
	    if(e.getValue()>best.getValue())
		best = e;
	}
	double value = (double) best.getValue()/group.numberOfIObjects();
		
	return new Measure("Group Purity", value, "Group Purity", best.getKey());
	
    }

    /**
     * partSplitValues should contain ordered values between 0 and 1 representing subparts of the entire group.
     * Purity is calculated for the corresponding subparts of the entire ordered group.
     * For instance, for a group consisting of 100 IObjects with partSplitValues [0.25, 0.5, 0.75],
     * the purity is calculated for the 25 top IObjects, the 50 top IObjects, and the 75 top IObjects
     * respectively.
     * 
     * For a part purity calculation 
     * consecutive values in partSplitValues should yield a difference of at least 1 IObject.
     * This means that part purity calculations on small groups might lead to surprising results.
     *  
     * @param group
     * @param grouping
     * @param partSplitValues
     * @return
     */
	
    public static MultiMeasure getGroupPartPurity(IObjectGroup group, IObjectGrouping grouping, float[] partSplitValues){

	HashMap<IObjectGroup, Integer> groupCount = new HashMap<IObjectGroup, Integer>();
	for(int j = 0; j < grouping.getGroups().size(); j++){
	    groupCount.put(grouping.getGroups().get(j), 0);
	}
		
	int[] parts = new int[partSplitValues.length];
	for(int i = 0; i < partSplitValues.length; i++){
	    parts[i] = (int) Math.floor(partSplitValues[i]*(group.numberOfIObjects()));
	}
	Vector<IObject> iovec = group.getOrdered();
		
	int p = 0;
	Measure temp = getGroupPurity(group, grouping);
	MultiMeasure mm = new MultiMeasure("Group Part Purity", temp.getValue(), "Group Part Purity", temp.getRelativeIObjectGroup());
	for(int i = 0; i < iovec.size(); i++){

	    IObject io = iovec.get(i);
	    IObjectGroup countGroup = grouping.inWhichIObjectGroup(io);
	    Integer count = groupCount.get(countGroup);
	    count++;
	    groupCount.put(countGroup, count);
			
	    if(p < parts.length && i+1 >= parts[p]){
			
		Set<Map.Entry<IObjectGroup, Integer>> countset = groupCount.entrySet();
		Iterator it = countset.iterator();
		Map.Entry<IObjectGroup, Integer> best = (Map.Entry<IObjectGroup, Integer>) it.next();
		while(it.hasNext()){
		    Map.Entry<IObjectGroup, Integer> e = (Map.Entry<IObjectGroup, Integer>) it.next();
		    if(e.getValue()>best.getValue())
			best = e;
		}
		double value = (double) best.getValue()/i;
		String name = partSplitValues[p]+"-purity";
		Measure m = new Measure(name, value, name, best.getKey());
		mm.addMeasure(m);
		p++;
	    }
	}
	return mm;
	
    }





}
