package infomat.evaluation;

import java.util.HashMap;
import java.util.Vector;
import java.io.*;

import infomat.vectorspace.*;
import infomat.vectorspace.similarity.*;
import infomat.vectorspace.ivector.*;

import mro.util.*;
import mro.util.measures.*;


/**
 * For calculating intrinsic quality measures per IObjectGroup.
 * 
 * @version          080424
 * @author           Magnus Rosell
 */
public class IntrinsicGroupEvaluation {

    //    private IMatrix matrix;
    //    private IObjectGroup group;
    //    private IObjectGrouping grouping;
    private ISimilarity similarity;

    public IntrinsicGroupEvaluation(
				    //				    IMatrix matrix, 
				    //				    IObjectGroup group, 
				    //				    IObjectGrouping grouping, 
				    ISimilarity similarity){
	//	this.matrix = matrix;
	//	this.group = group;
	//	this.grouping = grouping;
	this.similarity = similarity;
    }


    public double selfSimilarity(IObjectGroup group){
	//	System.out.println("selfSimilarity() "+group);

	if(similarity instanceof SparseISimilarity){
	    //	    System.out.println("selfSimilarity() sparse");
	    SparseISimilarity sis = (SparseISimilarity) similarity;
	    IVector<IVectorIndex> ivec = 
		sis.getGroupRepresentation(group);
	    return sis.sim(ivec, ivec);
	}else if(similarity instanceof DenseISimilarity){
	    //	    System.out.println("selfSimilarity() dense");
	    DenseISimilarity dis = (DenseISimilarity) similarity;
	    float[] repr = 
		dis.getUnnormalizedGroupRepresentation(group);
	    //	    System.out.println("selfSimilarity() dense repr");
	    if(repr!=null)
		return dis.sim(repr, repr);
	}
	return Double.MAX_VALUE;
    }

    /**
     * Uses getGroupRepresentation() in the ISimilarity.
     * Treats it as if it is a centroid, to calcualate
     * the average weight of all IObject:s to all IObject:s
     * in the IObjectGroup. The similarity of all IObject:s 
     * to them selves are subtracted, so it becomes the
     * the average similarity BETWEEN all IObject:s.
     */
    /*
    public double selfSimilarity(IObjectGroup group){
	
	// The similarity of the objects to themselves.
	double ownSim = 0.0;
	Vector<IObject> ios = group.getOrdered();
	for(int i=0; i<ios.size(); i++){
	    HashMap<IObject, Float> map = 
		similarity.getRepresentation(ios.elementAt(i));
	    ownSim += similarity.sim(map, map);
	}

	HashMap<IObject, Float> iofMap = 
	    similarity.getGroupRepresentation(group);
	double fullSelfSim = similarity.sim(iofMap, iofMap);
	int groupSize = group.numberOfIObjects();

	return 
	    (Math.pow(groupSize, 2) * fullSelfSim - ownSim) / 
	    (Math.pow(groupSize, 2) - groupSize);
    }
    */

    public double groupGroupingSimilarity(IObjectGroup group, 
					  IObjectGrouping grouping){
	if(similarity instanceof SparseISimilarity){
	    SparseISimilarity sis = (SparseISimilarity) similarity;
	    IVector<IVectorIndex> ivec = 
		sis.getGroupRepresentation(group);
	    IVector<IVectorIndex> ivec2 = 
		sis.getGroupRepresentation(grouping);
	    return sis.sim(ivec, ivec2);
	}else if(similarity instanceof DenseISimilarity){
	    DenseISimilarity dis = (DenseISimilarity) similarity;
	    float[] repr = 
		dis.getUnnormalizedGroupRepresentation(group);
	    float[] repr2 = 
		dis.getUnnormalizedGroupRepresentation(grouping);
	    return dis.sim(repr, repr2);
	}
	return -Double.MAX_VALUE;
    }


    /**
     * Uses getGroupRepresentation() in the ISimilarity.
     * Treats it as if it is a centroid, to calcualate
     * the average weight of all IObject:s of the IObjectGroup
     * to all IObject:s in the IObjectGrouping. 
     * The similarity of all IObject:s 
     * to them selves are subtracted, so it becomes the
     * the average similarity BETWEEN all IObject:s in the group
     * to the all IObject:s in the grouping.
     */
    /*
    public double groupGroupingSimilarity(IObjectGroup group, 
					  IObjectGrouping grouping){

	// The similarity of the objects to themselves.
	double ownSim = 0.0;
	Vector<IObject> ios = group.getOrdered();
	for(int i=0; i<ios.size(); i++){
	    IObject io = ios.elementAt(i);
	    if(grouping.contains(io)){
		HashMap<IObject, Float> map = 
		    similarity.getRepresentation(io);
		ownSim += similarity.sim(map, map);
	    }
	}

	HashMap<IObject, Float> iofMap = 
	    similarity.getGroupRepresentation(group);
	HashMap<IObject, Float> iofMap2 = 
	    similarity.getGroupRepresentation(grouping);
	double fullSimilarity = similarity.sim(iofMap, iofMap2);
	int groupSize = group.numberOfIObjects();
	int groupingSize = grouping.numberOfIObjects();
	
	return (groupSize*groupingSize*fullSimilarity - ownSim) / 
	    (groupSize*groupingSize - groupSize);
    }
    */


    public double weightedAvgSelfSimilarity(Vector<IObjectGroup> groups){
	double sim = 0.0;
	int tot = 0;
	for(int i=0; i<groups.size(); i++){
	    IObjectGroup group = groups.elementAt(i);
	    sim += group.numberOfIObjects() * selfSimilarity(group);
	    tot += group.numberOfIObjects();
	}
	sim /= tot;

	return sim;
    }

    public double weightedAvgSelfSimilarity(IObjectGrouping grouping){
	double sim = 0.0;
	Vector<IObjectGroup> groups = grouping.getGroups();
	for(int i=0; i<groups.size(); i++){
	    IObjectGroup group = groups.elementAt(i);
	    sim += group.numberOfIObjects() * selfSimilarity(group);
	    //	    sim += Math.pow(group.numberOfIObjects(),2) * selfSimilarity(group);
	}
	sim /= grouping.numberOfIObjects();
	//	sim /= Math.pow(grouping.numberOfIObjects(), 2);

	return sim;
    }

    public double weightedAvgGroupGroupingSimilarity(IObjectGrouping grouping){

	double sim = 0.0;
	Vector<IObjectGroup> groups = grouping.getGroups();
	for(int i=0; i<groups.size(); i++){
	    IObjectGroup group = groups.elementAt(i);
	    sim += group.numberOfIObjects() * 
		groupGroupingSimilarity(group, grouping);
	}
	sim /= grouping.numberOfIObjects();

	return sim;
    }



    public Measure selfSimilarityMeasure(IObjectGroup group){
	Measure m = new Measure("Self Similarity", selfSimilarity(group), "The similarity of the group with itself.");
	return m;
    }

    public Measure groupGroupingSimilarityMeasure(IObjectGroup group, 
						  IObjectGrouping grouping){
	Measure m = new Measure("Grouping Similarity", groupGroupingSimilarity(group, grouping), "The similarity to the whole grouping considered one group.");
	return m;
    }


    public HashMap<IObjectGroup, Measure> getSelfSimilarityMap(Vector<IObjectGroup> groups){
	HashMap<IObjectGroup, Measure> iogmeasure = 
	    new HashMap<IObjectGroup, Measure>();
	for(int i = 0; i < groups.size(); i++){
	    IObjectGroup iog = groups.get(i);
	    Measure m = selfSimilarityMeasure(iog);
	    iogmeasure.put(iog, m);
	}

	return iogmeasure;
    }

    public HashMap<IObjectGroup, Measure> getGroupGroupingSimilarityMap(Vector<IObjectGroup> groups, IObjectGrouping grouping){
	HashMap<IObjectGroup, Measure> iogmeasure = 
	    new HashMap<IObjectGroup, Measure>();
	for(int i = 0; i < groups.size(); i++){
	    IObjectGroup iog = groups.get(i);
	    Measure m = groupGroupingSimilarityMeasure(iog, grouping);
	    iogmeasure.put(iog, m);
	}

	return iogmeasure;
    }


}
