package infomat.io;

import infomat.vectorspace.*;
import infomat.vectorspace.ivector.*;

import java.io.*;
import java.util.*; 
import java.util.zip.*;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;

/**
 * For loading and saving a IVector<T>. Only loads those IObjects 
 * that are in the IObjectGroup provided!
 * 
 * @version          080704
 * @author           Magnus Rosell
 */
public class IVectorIO<T extends IVectorIndex>{

    public IVector<T> load(IObjectGroup group, File file) {
	try {
	    BufferedReader in = 
		new BufferedReader(new FileReader(file));
	    InputSource source = new InputSource(in);
	    return load(source, group);
	}
	catch(IOException ioE) {
	    System.out.println(ioE);
	    ioE.printStackTrace();
	}
	return null;
    }

    public IVector<T> loadCompressed(IObjectGroup group, File file) {
	try {
	    // Create a ZipInputStream to read the zip file            
	    ZipInputStream zis = new ZipInputStream(new BufferedInputStream(new FileInputStream(file)));
	    InputSource source = new InputSource(zis);
	    ZipEntry entry;
	    if((entry = zis.getNextEntry()) != null) {
		if(!entry.isDirectory()) {
		    return load(source, group);
		}
	    }
	    zis.close();
	}
	catch(IOException ioE) {
	    System.out.println(ioE);
	    ioE.printStackTrace();
	}
	return null;

    }

    private IVector<T> load(InputSource source, IObjectGroup group) {
	try {
	    // Create parser+contentHandler and read the XML document
	    XMLReader parser = XMLReaderFactory.createXMLReader();
	    IVectorOfIObjectsHandler<T> handler = new IVectorOfIObjectsHandler<T>(group);
	    parser.setContentHandler(handler);
	    parser.parse(source);

	    IVector<T> vec = handler.getIVector();
	    return vec;
	} catch (Exception e) {
	    System.err.println(e);
	    e.printStackTrace();
	}
	return null;
    }

    public void save(IVector<T> vec, File file){
	try {
	    PrintWriter out =
		new PrintWriter(new BufferedWriter(new BufferedWriter(new FileWriter(file))));
	    save(out, vec);
	    out.close();
	}
	catch(IOException ioE) {
	    System.out.println(ioE);
	    ioE.printStackTrace();
	}
    }


    public void saveCompressed(IVector<T> vec, String fileName) { 
	try {
	    // Create zip compressed output stream, zip file entry and PrintWriter
	    ZipOutputStream targetStream = new ZipOutputStream(new FileOutputStream(fileName+".xml.zip"));
	    File file = new File(fileName);                        
	    targetStream.putNextEntry(new ZipEntry(file.getName() + ".xml"));
	    PrintWriter out = new PrintWriter(targetStream);
	    save(out, vec);            
	    // Close compressed output stream and PrintWriter
	    out.flush();
	    targetStream.closeEntry();
	    targetStream.close();
	    targetStream = null;
	    out.close();
	}
	catch (java.io.IOException ioE) {
	    System.out.println(ioE);
	    ioE.printStackTrace();
	}
    }




    public void printDTD(PrintWriter out) {
	out.println("<?xml version='1.0' encoding='ISO-8859-1' standalone=\"yes\"?>");

	out.println("<!DOCTYPE IVector [");

	printPartDTD(out);

	out.println("]>");
    }




    public void printPartDTD(PrintWriter out) {
	out.println("<!-- IVector -->");
	out.println("<!ELEMENT IVector (IVecEl+)>");
	out.println("<!ELEMENT IVecEl (IVStr, IVVal)> <!-- An IVector Element -->");
	out.println("<!ELEMENT IVStr (#PCDATA)> <!-- A string representing the IVector Index -->");
	out.println("<!ELEMENT IVVal (#PCDATA)> <!-- Order value -->");
    }





    public void save(PrintWriter out, IVector<T> vec) {
	printDTD(out);
	savePart(out, vec);
    }



    public void savePart(PrintWriter out, IVector<T> vec) {
	out.println("<IVector>");
	Vector<T> ios = vec.getOrderedVector();
	for(int i=0; i<ios.size(); i++){
	    T io = ios.elementAt(i);
	    out.println("<IVecEl>");
	    out.println("<IVStr>"+io.getString()+"</IVStr>");
	    out.println("<IVVal>"+vec.getValue(io)+"</IVVal>");
	    out.println("</IVecEl>");
	}
	out.println("</IVector>");
	out.flush();
    }




    private class IVectorOfIObjectsHandler<T extends IVectorIndex> extends DefaultHandler {
	private StringBuffer buffer;

	private IVector vector;
	private IObjectGroup group;
	private HashMap<String, IObject> strIOMap;

	private String ioString;
	private float val;
		
	public IVectorOfIObjectsHandler(IObjectGroup group){
	    this.group = group;
	    vector = new IVector<T>();

	    strIOMap = new HashMap<String, IObject>();
	    Vector<IObject> ios = group.getOrdered();
	    for(int i=0; i<ios.size(); i++){
		IObject io = ios.elementAt(i);
		strIOMap.put(io.getString(), io);
	    }
	}

	public IVector<T> getIVector(){
	    return vector;
	}

	public void startElement (String namespaceURI, String localName,
				  String qualifiedName, Attributes atts) {
	    buffer = new StringBuffer();
	}

	public void endElement (String namespaceURI, String localName, String qualifiedName) {

	    if(qualifiedName.equals("IVStr")){
		ioString = buffer.toString().trim();
	    }else if(qualifiedName.equals("IVVal")){
		val = Float.parseFloat(buffer.toString().trim());
	    }else if(qualifiedName.equals("IVecEl")){
		IObject io = strIOMap.get(ioString);
		if(io!=null){
		    vector.add(io, val);
		}
	    }			
	}

	public void characters (char ch[], int start, int length) {
	    if(buffer != null)
		buffer.append(ch,start,length);
	}      
    }




}
